/*
	Ejemplo de reproduccin de sonidos muestreados en la Gameboy Advance mediante doble buffer.

	Avelino Herrera Morales, 2005

	Basado en los ejemplos propuestos en http://belogic.com/gba
	He utilizado la toolchain descrita en http://linux.gbadev.org basada en binutils/gcc/newlib
	Cuidado con el fichero crt0.o, ya que debe tener habilitadas las interrupciones (http://devrs.com/gba).
*/

#include <stdlib.h>
#include <string.h>
#include "gba.h"



#define  SAMPLE_RATE  16384

#define  BUFFER_SIZE    2048
#define  BUFFER_SIZE_2  (BUFFER_SIZE / 2)
#define  BUFFER_SIZE_4  (BUFFER_SIZE / 4)
#define  BUFFER_SIZE_8  (BUFFER_SIZE / 8)


/* sample a reproducir en modo doble buffer */
extern const u32 buffer_ritmo[];
int buffer_ritmo_size = 35747;
/* ---------------------------------------- */


u32 buffer[BUFFER_SIZE_4];  /* la muestra, 8 bits con signo a 16KHz */
u32 *buffer_middle = &(buffer[BUFFER_SIZE_8]);  /* puntero a la mitad del buffer */

void InterruptProcess(void) __attribute__ ((section(".iwram"))); /* La ISR debe declararse as obligatoriamente */
void dma_play_sound(void);
void dma_stop_sound(void);

int sub_buffer = 0;
int last_written = 1;


u32 temp_buffer[BUFFER_SIZE_8];

 
void InterruptProcess(void) {
	/* se acab el sample */
	if (sub_buffer == 0) {
		/* Termin de reproducirse la primera mitad del buffer */
		REG_DMA1CNT_H = 0; /* paramos la transferencia por DMA */
		REG_DMA1SAD = (unsigned long) buffer_middle; /* nuevo buffer origen para el DMA1 (la segunda mitad del buffer) */
		REG_DMA1CNT_H = 0xb600; /* reactivamos el DMA1 + FIFO + 32 bits + con repeticin */
		sub_buffer = 1;
	}
	else {
		/* termin de reproducirse la segunda mitad del buffer */
		REG_DMA1CNT_H = 0; /* paramos la transferencia por DMA */
		REG_DMA1SAD = (unsigned long) buffer; /* nuevo buffer origen para el DMA1 (la primera mitad del buffer) */
		REG_DMA1CNT_H = 0xb600; /* reactivamos el DMA1 + FIFO + 32 bits + con repeticin */
		sub_buffer = 0;
	}
	/* borramos siempre el flag de interrupcin */
	REG_IF |= REG_IF;
}


void dma_stop_sound(void) {
	REG_TM0CNT_H = 0; /* deshabilitamos el timer 0 */
	REG_DMA1CNT_H = 0; /* paramos la transferencia por DMA */
}

 
void dma_play_sound(void) {
	REG_TM0CNT_H = 0; /* deshabilitamos el timer 0*/
	/* reproducimos un sonido mono a 16KHz en modo DMA */
	/* utilizamos el timer 0 como generador del perodo de muestreo */
	/* y el timer 1 en cascada para contar los samples que se van enviando a la FIFO del chip de sonido */
	REG_SOUNDCNT_L = 0;
	REG_SOUNDCNT_H = 0x0b0f; /* Direct Sound A&B + reiniciar la FIFO + timer0 + volumen mximo a los canales L y R */
	REG_SOUNDCNT_X = 0x0080; /* activamos el chip de sonido */
	
	REG_DMA1SAD = (unsigned long) buffer; /* buffer origen para el DMA1 */
	REG_DMA1DAD = 0x040000a0; /* buffer destino = direccin del puerto FIFO del chip de sonido */
	REG_DMA1CNT_H = 0xb600; /* activamos el DMA1 + FIFO + 32 bits + con repeticin */
	
	REG_TM1CNT_L = 0xffff - BUFFER_SIZE_2; /* 0xffff - nmero de muestras (la primera mitad del buffer) */
	REG_TM1CNT_H = 0xC4; /* habilitamos el timer 1 + irq y en cascada con el timer 0 */

	REG_IE = 0x10; /* habilitar interrupcin al desbordarse el timer 1 */
	REG_IME = 1; /* habilitar las interrupciones */
	
	/* La frmula para obtener el valor del timer 0: 0xFFFF - round(frecCPU / frecMuestreo) siendo frecCPU = 16MHz */
	//REG_TM0CNT_L = 0xFBE8; /* 16kHz */
	REG_TM0CNT_L = 0xFBE8; /* 16kHz */
	REG_TM0CNT_H = 0x0080; /* habilitar timer 0 a la frecuencia de la CPU */
	/* indicamos que estamos reproduciendo la primera mitad del buffer */
	sub_buffer = 0;
}


void write_sound(u32 *data) {
	if (last_written == 0) {
		while (sub_buffer == 1)
			;
		/* escribimos en la segunda mitad del buffer */
		memcpy(buffer_middle, data, BUFFER_SIZE_2);
		last_written = 1;
	}
	else {
		while (sub_buffer == 0)
			;
		/* escribimos en la primera mitad del buffer */
		memcpy(buffer, data, BUFFER_SIZE_2);
		last_written = 0;
	}
}


int AgbMain(void) {
	int i, j;

	dma_play_sound();
	j = 0;
	while (1) {
		for (i = 0; i < BUFFER_SIZE_8; i++, j++) {
			if (j >= (buffer_ritmo_size / 4))
				j = 0;
			temp_buffer[i] = buffer_ritmo[j];
		}
		write_sound(temp_buffer);
	}
}

